#include <stdarg.h>
#include "ntddk.h"
#include "trace.h"
#include "wdf.h"
#include "prototypes.h"
#include "usbdi.h"
#include "wdfusb.h"
#include "initguid.h"
 
//    
#define BULK_READ_ENDPOINT_ADDR		0x85
//    
#define BULK_WRITE_ENDPOINT_ADDR	0x06
 
//-   (       )
// {1B2E6756-0344-449e-8B5B-BD57F582766B}
DEFINE_GUID(GUID_DEVINTERFACE_FLIRONEGEN2_FRAME,0x1b2e6756, 0x344, 0x449e, 0x8b, 0x5b, 0xbd, 0x57, 0xf5, 0x82, 0x76, 0x6b);

//   (           )
typedef struct _DEVICE_CONTEXT 
{
 WDFUSBDEVICE UsbDevice;// 
 WDFUSBINTERFACE UsbInterface;// 
 WDFUSBPIPE BulkReadPipe;//  
 WDFUSBPIPE BulkWritePipe;//  
} DEVICE_CONTEXT,*PDEVICE_CONTEXT;
 
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(DEVICE_CONTEXT,GetDeviceContext)
 
//----------------------------------------------------------------------------------------------------
//,    
//DriverObject -   -,   . 
//RegistryPath -      Unicode    ,   .
//----------------------------------------------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,IN PUNICODE_STRING RegistryPath)
{
 NTSTATUS status;// 
 WDF_DRIVER_CONFIG config;// 
 WDF_OBJECT_ATTRIBUTES attributes;// 
  
 WPP_INIT_TRACING(DriverObject,RegistryPath);//  ( )
  
 KdPrint(("FlirOneGen2_Frame: DriverEntry.\n"));//     
 WDF_DRIVER_CONFIG_INIT(&config,EvtDeviceAdd);//  
  
 // 
 WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
 attributes.EvtCleanupCallback=EvtDriverContextCleanup;
  
 //  
 status=WdfDriverCreate(DriverObject,RegistryPath,&attributes,&config,WDF_NO_HANDLE);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_Frame: WdfDriverCreate failed 0x%x\n",status));
  WPP_CLEANUP(DriverObject);
  return(status);
 } 
 return(status);
}
//----------------------------------------------------------------------------------------------------
//   
//----------------------------------------------------------------------------------------------------
VOID EvtDriverContextCleanup(IN WDFDRIVER Driver)
{
 WPP_CLEANUP(WdfDriverWdmGetDriverObject(Driver));
}
 
//----------------------------------------------------------------------------------------------------
//,    
//---------------------------------------------------------------------------------------------------- 
NTSTATUS EvtDeviceAdd(IN WDFDRIVER Driver,IN PWDFDEVICE_INIT DeviceInit)
{
 NTSTATUS status;// 
 WDFDEVICE device;//
 WDF_OBJECT_ATTRIBUTES attributes;//  (   )
 PDEVICE_CONTEXT pDevContext;// 
 WDF_PNPPOWER_EVENT_CALLBACKS pnpPowerCallbacks;//  
 WDF_IO_QUEUE_CONFIG ioQueueConfig;//   
  
 UNREFERENCED_PARAMETER(Driver);//,       
  
 //     
 WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnpPowerCallbacks);
 pnpPowerCallbacks.EvtDevicePrepareHardware=EvtDevicePrepareHardware;//   
 WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit,&pnpPowerCallbacks);//   
 WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes,DEVICE_CONTEXT);
  
 status=WdfDeviceCreate(&DeviceInit,&attributes,&device);// 
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_Frame: WdfDeviceCreate failed 0x%x\n",status));
  return(status);
 }
  
 pDevContext=GetDeviceContext(device);//  
 //      (      )
 status=WdfDeviceCreateDeviceInterface(device,(LPGUID)&GUID_DEVINTERFACE_FLIRONEGEN2_FRAME,NULL);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_Frame: WdfDeviceCreateDeviceInterface failed 0x%x\n",status));
  return(status);
 }
 //       
 WDF_IO_QUEUE_CONFIG_INIT_DEFAULT_QUEUE(&ioQueueConfig,WdfIoQueueDispatchParallel);
 //ioQueueConfig.EvtIoDeviceControl=EvtIoDeviceControl;
 ioQueueConfig.EvtIoRead=EvtIoRead;
 ioQueueConfig.EvtIoWrite=EvtIoWrite;
  
 status=WdfIoQueueCreate(device,&ioQueueConfig,WDF_NO_OBJECT_ATTRIBUTES,WDF_NO_HANDLE);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_Frame: WdfIoQueueCreate failed 0x%x\n",status));
  return(status);
 }
  
 return(status);
}
 
//----------------------------------------------------------------------------------------------------
//,    
//----------------------------------------------------------------------------------------------------
NTSTATUS EvtDevicePrepareHardware(IN WDFDEVICE Device,IN WDFCMRESLIST ResourceList,IN WDFCMRESLIST ResourceListTranslated)
{
 long n;
 NTSTATUS status;// 
 PDEVICE_CONTEXT pDeviceContext;// 
 WDF_USB_DEVICE_SELECT_CONFIG_PARAMS configParams;// 
 WDFUSBPIPE pipe;//
 long ep_index=0;//  
 WDF_USB_PIPE_INFORMATION pipeConfig;// 
  
 UNREFERENCED_PARAMETER(ResourceList);//,       
 UNREFERENCED_PARAMETER(ResourceListTranslated);//,       
  
 pDeviceContext=GetDeviceContext(Device);//  
  
 //  USB-,      
 if (pDeviceContext->UsbDevice==NULL)//   USB-
 {
  status=WdfUsbTargetDeviceCreate(Device,WDF_NO_OBJECT_ATTRIBUTES,&pDeviceContext->UsbDevice);
  if (!NT_SUCCESS(status))// 
  {
   KdPrint(("FlirOneGen2_Frame: WdfUsbTargetDeviceCreate failed 0x%x\n",status)); 
   return(status);
  }
 }
  
 //   
 WDF_USB_DEVICE_SELECT_CONFIG_PARAMS_INIT_SINGLE_INTERFACE(&configParams);
 //     USB
 status=WdfUsbTargetDeviceSelectConfig(pDeviceContext->UsbDevice,WDF_NO_OBJECT_ATTRIBUTES,&configParams);
 if(!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_Frame: WdfUsbTargetDeviceSelectConfig failed 0x%x\n", status));
  return(status);
 }
 pDeviceContext->UsbInterface=configParams.Types.SingleInterface.ConfiguredUsbInterface;
 // 2 :     
 pDeviceContext->BulkReadPipe=NULL;
 pDeviceContext->BulkWritePipe=NULL;
 WDF_USB_PIPE_INFORMATION_INIT(&pipeConfig);
 while(1)
 {
  pipe=WdfUsbInterfaceGetConfiguredPipe(pDeviceContext->UsbInterface,ep_index,&pipeConfig);
  if (pipe==NULL) break;//   
  KdPrint(("FlirOneGen2_Frame: EndPoint addr:0x%02x\n",pipeConfig.EndpointAddress));
  if (pipeConfig.EndpointAddress==BULK_READ_ENDPOINT_ADDR) pDeviceContext->BulkReadPipe=pipe;
  if (pipeConfig.EndpointAddress==BULK_WRITE_ENDPOINT_ADDR) pDeviceContext->BulkWritePipe=pipe;
  ep_index++;
 }
 //  
 if (pDeviceContext->BulkReadPipe!=NULL)
 {
  KdPrint(("FlirOneGen2_Frame: BulkReadPipe is OK.\n"));
  WdfUsbTargetPipeSetNoMaximumPacketSizeCheck(pDeviceContext->BulkReadPipe);//      
 }
 else KdPrint(("FlirOneGen2_Frame: BulkReadPipe is NULL!\n"));
 
 //  
 if (pDeviceContext->BulkWritePipe!=NULL)
 {
  KdPrint(("FlirOneGen2_Frame: BulkWritePipe is OK.\n"));
  WdfUsbTargetPipeSetNoMaximumPacketSizeCheck(pDeviceContext->BulkWritePipe);//     
 }
 else KdPrint(("FlirOneGen2_Frame: BulkWritePipe is NULL!\n"));
 return(status);
}

//----------------------------------------------------------------------------------------------------
//  
//Queue-.
//Request- .
//Length-   
//----------------------------------------------------------------------------------------------------
VOID EvtIoRead(IN WDFQUEUE Queue,IN WDFREQUEST Request,IN size_t Length)
{
 NTSTATUS status;// 
 WDFUSBPIPE pipe;//
 WDFMEMORY reqMemory;
 PDEVICE_CONTEXT pDeviceContext;// 
 BOOLEAN ret;
  
 UNREFERENCED_PARAMETER(Length);//,       
  
 KdPrint(("FlirOneGen2_Frame: EvtIoRead.\n"));//     

 pDeviceContext=GetDeviceContext(WdfIoQueueGetDevice(Queue));//  
  
 pipe=pDeviceContext->BulkReadPipe;//   
  
 if (pipe==NULL)
 {
  KdPrint(("FlirOneGen2_Frame: Pipe is NULL!\n"));//     
  status=STATUS_SUCCESS;
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }

 status=WdfRequestRetrieveOutputMemory(Request,&reqMemory);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_Frame: WdfRequestRetrieveOutputMemory failed 0x%x", status));
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
 //   
 status=WdfUsbTargetPipeFormatRequestForRead(pipe,Request,reqMemory,NULL);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_Frame: WdfUsbTargetPipeFormatRequestForRead failed 0x%x", status));
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
 //  
 WdfRequestSetCompletionRoutine(Request,EvtRequestReadCompletionRoutine,pipe);
  
 ret=WdfRequestSend(Request,WdfUsbTargetPipeGetIoTarget(pipe),WDF_NO_SEND_OPTIONS);
 if (ret==FALSE) 
 {
  KdPrint(("FlirOneGen2_Frame: WdfRequestSend failed 0x%x", status));
  status=WdfRequestGetStatus(Request);
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return; 
 } 
 else return;
 KdPrint(("FlirOneGen2_Frame: Ok."));
 WdfRequestCompleteWithInformation(Request,status,0);// 
 return;
}
 
//----------------------------------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------------------------------
VOID EvtRequestReadCompletionRoutine(IN WDFREQUEST Request,IN WDFIOTARGET Target,PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,IN WDFCONTEXT Context)
{
 NTSTATUS status;// 
 size_t bytesRead=0;//  
 PWDF_USB_REQUEST_COMPLETION_PARAMS usbCompletionParams;
  
 UNREFERENCED_PARAMETER(Target);//,       
 UNREFERENCED_PARAMETER(Context);//,       
  
 status=CompletionParams->IoStatus.Status;
  
 usbCompletionParams=CompletionParams->Parameters.Usb.Completion;
 bytesRead=usbCompletionParams->Parameters.PipeRead.Length;
  
 if (NT_SUCCESS(status))//  
 {
  KdPrint(("FlirOneGen2_Frame: Number of bytes read: %I64d\n", (INT64)bytesRead));
 } 
 else 
 {
  KdPrint(("FlirOneGen2_Frame: Read failed - request status 0x%x UsbdStatus 0x%x\n",
  status, usbCompletionParams->UsbdStatus));
 }
 WdfRequestCompleteWithInformation(Request,status,bytesRead);// 
 return;
}
//----------------------------------------------------------------------------------------------------
//  
//Queue-.
//Request- .
//Length-   
//----------------------------------------------------------------------------------------------------
VOID EvtIoWrite(IN WDFQUEUE Queue,IN WDFREQUEST Request,IN size_t Length)
{
 NTSTATUS status;// 
 WDFUSBPIPE pipe;//
 WDFMEMORY reqMemory;
 PDEVICE_CONTEXT pDeviceContext;// 
 BOOLEAN ret;
  
 UNREFERENCED_PARAMETER(Length);//,       
  
 pDeviceContext=GetDeviceContext(WdfIoQueueGetDevice(Queue));//  
  
 pipe=pDeviceContext->BulkWritePipe;//   
  
 status=WdfRequestRetrieveInputMemory(Request,&reqMemory);
 if (!NT_SUCCESS(status))// 
 {
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
 //  
 status=WdfUsbTargetPipeFormatRequestForWrite(pipe,Request,reqMemory,NULL);
 if (!NT_SUCCESS(status))// 
 {
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
  
 WdfRequestSetCompletionRoutine(Request,EvtRequestWriteCompletionRoutine,pipe);
  
 ret=WdfRequestSend(Request,WdfUsbTargetPipeGetIoTarget(pipe),WDF_NO_SEND_OPTIONS);
 if (ret==FALSE)//
 {
  status=WdfRequestGetStatus(Request);
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 } 
 else return;
  
 WdfRequestCompleteWithInformation(Request,status,0);// 
 return;
}
//----------------------------------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------------------------------
VOID EvtRequestWriteCompletionRoutine(IN WDFREQUEST Request,IN WDFIOTARGET Target,PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,IN WDFCONTEXT Context)
{
 NTSTATUS status;
 size_t bytesWritten=0;//  
 PWDF_USB_REQUEST_COMPLETION_PARAMS usbCompletionParams;
  
 UNREFERENCED_PARAMETER(Target);//,       
 UNREFERENCED_PARAMETER(Context);//,       
  
 status=CompletionParams->IoStatus.Status;
  
 usbCompletionParams=CompletionParams->Parameters.Usb.Completion;
  
 bytesWritten=usbCompletionParams->Parameters.PipeWrite.Length;
 if (NT_SUCCESS(status))//  
 {
  KdPrint(("FlirOneGen2_Frame: Number of bytes written: %I64d\n",(INT64)bytesWritten));
 }
 else// 
 {
  KdPrint(("FlirOneGen2_Frame: Write failed: request Status 0x%x UsbdStatus 0x%x\n",
  status, usbCompletionParams->UsbdStatus));
 }
 WdfRequestCompleteWithInformation(Request,status,bytesWritten);// 
 return;
}
