#include <stdarg.h>
#include "ntddk.h"
#include "trace.h"
#include "wdf.h"
#include "prototypes.h"
#include "usbdi.h"
#include "wdfusb.h"
#include "initguid.h"
 
//    
#define BULK_READ_ENDPOINT_ADDR		0x81
//    
#define BULK_WRITE_ENDPOINT_ADDR	0x02
 
// 
//  Frame
#define IOCTL_FLIR_ONE_GEN2_FRAME_STOP		1
//  Frame
#define IOCTL_FLIR_ONE_GEN2_FRAME_START		2
//  FileIO
#define IOCTL_FLIR_ONE_GEN2_FILEIO_STOP		4
//  FileIO
#define IOCTL_FLIR_ONE_GEN2_FILEIO_START	8

 
//-   (       )
// {41A83025-9DC8-4e75-8705-1209222FDC17}
DEFINE_GUID(GUID_DEVINTERFACE_FLIRONEGEN2_IAP,0x41a83025, 0x9dc8, 0x4e75, 0x87, 0x5, 0x12, 0x9, 0x22, 0x2f, 0xdc, 0x17);

//   (           )
typedef struct _DEVICE_CONTEXT 
{
 WDFUSBDEVICE UsbDevice;// 
 WDFUSBINTERFACE UsbInterface;// 
 WDFUSBPIPE BulkReadPipe;//  
 WDFUSBPIPE BulkWritePipe;//  
} DEVICE_CONTEXT,*PDEVICE_CONTEXT;
 
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(DEVICE_CONTEXT,GetDeviceContext)
 
//----------------------------------------------------------------------------------------------------
//,    
//DriverObject -   -,   . 
//RegistryPath -      Unicode    ,   .
//----------------------------------------------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,IN PUNICODE_STRING RegistryPath)
{
 NTSTATUS status;// 
 WDF_DRIVER_CONFIG config;// 
 WDF_OBJECT_ATTRIBUTES attributes;// 
  
 WPP_INIT_TRACING(DriverObject,RegistryPath);//  ( )
  
 KdPrint(("FlirOneGen2_iAP: DriverEntry.\n"));//     
 WDF_DRIVER_CONFIG_INIT(&config,EvtDeviceAdd);//  
  
 // 
 WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
 attributes.EvtCleanupCallback=EvtDriverContextCleanup;
  
 //  
 status=WdfDriverCreate(DriverObject,RegistryPath,&attributes,&config,WDF_NO_HANDLE);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_iAP: WdfDriverCreate failed 0x%x\n",status));
  WPP_CLEANUP(DriverObject);
  return(status);
 } 
 return(status);
}
//----------------------------------------------------------------------------------------------------
//   
//----------------------------------------------------------------------------------------------------
VOID EvtDriverContextCleanup(IN WDFDRIVER Driver)
{
 WPP_CLEANUP(WdfDriverWdmGetDriverObject(Driver));
}
 
//----------------------------------------------------------------------------------------------------
//,    
//---------------------------------------------------------------------------------------------------- 
NTSTATUS EvtDeviceAdd(IN WDFDRIVER Driver,IN PWDFDEVICE_INIT DeviceInit)
{
 NTSTATUS status;// 
 WDFDEVICE device;//
 WDF_OBJECT_ATTRIBUTES attributes;//  (   )
 PDEVICE_CONTEXT pDevContext;// 
 WDF_PNPPOWER_EVENT_CALLBACKS pnpPowerCallbacks;//  
 WDF_IO_QUEUE_CONFIG ioQueueConfig;//   
  
 UNREFERENCED_PARAMETER(Driver);//,       
  
 KdPrint(("FlirOneGen2_iAP: EvtDeviceAdd\n")); 

 //     
 WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnpPowerCallbacks);
 pnpPowerCallbacks.EvtDevicePrepareHardware=EvtDevicePrepareHardware;//   
 WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit,&pnpPowerCallbacks);//   
 WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes,DEVICE_CONTEXT);
  
 status=WdfDeviceCreate(&DeviceInit,&attributes,&device);// 
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_iAP: WdfDeviceCreate failed 0x%x\n",status));
  return(status);
 }
  
 pDevContext=GetDeviceContext(device);//  
 //      (      )
 status=WdfDeviceCreateDeviceInterface(device,(LPGUID)&GUID_DEVINTERFACE_FLIRONEGEN2_IAP,NULL);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_iAP: WdfDeviceCreateDeviceInterface failed 0x%x\n",status));
  return(status);
 }
 //       
 WDF_IO_QUEUE_CONFIG_INIT_DEFAULT_QUEUE(&ioQueueConfig,WdfIoQueueDispatchParallel);
 ioQueueConfig.EvtIoDeviceControl=EvtIoDeviceControl;
 ioQueueConfig.EvtIoRead=EvtIoRead;
 ioQueueConfig.EvtIoWrite=EvtIoWrite;
  
 status=WdfIoQueueCreate(device,&ioQueueConfig,WDF_NO_OBJECT_ATTRIBUTES,WDF_NO_HANDLE);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_iAP: WdfIoQueueCreate failed 0x%x\n",status));
  return(status);
 }
  
 return(status);
}
 
//----------------------------------------------------------------------------------------------------
//,    
//----------------------------------------------------------------------------------------------------
NTSTATUS EvtDevicePrepareHardware(IN WDFDEVICE Device,IN WDFCMRESLIST ResourceList,IN WDFCMRESLIST ResourceListTranslated)
{
 long n;
 NTSTATUS status;// 
 PDEVICE_CONTEXT pDeviceContext;// 
 WDF_USB_DEVICE_SELECT_CONFIG_PARAMS configParams;// 
 UCHAR numInterfaces;
 WDFUSBPIPE pipe;//
 long ep_index=0;//  
 WDF_USB_PIPE_INFORMATION pipeConfig;// 
 
 UNREFERENCED_PARAMETER(ResourceList);//,       
 UNREFERENCED_PARAMETER(ResourceListTranslated);//,       
 //PWDF_USB_INTERFACE_SETTING_PAIR settingPairs;
 
 KdPrint(("FlirOneGen2_iAP: EvtDevicePrepareHardware\n")); 

 pDeviceContext=GetDeviceContext(Device);//  
  
 //  USB-,      
 if (pDeviceContext->UsbDevice==NULL)//   USB-
 {
  status=WdfUsbTargetDeviceCreate(Device,WDF_NO_OBJECT_ATTRIBUTES,&pDeviceContext->UsbDevice);
  if (!NT_SUCCESS(status))// 
  {
   KdPrint(("FlirOneGen2_iAP: WdfUsbTargetDeviceCreate failed 0x%x\n",status)); 
   return(status);
  }
 }  
 //   

 numInterfaces=WdfUsbTargetDeviceGetNumInterfaces(pDeviceContext->UsbDevice);
 //KdPrint(("FlirOneGen2_iAP: Interfaces:0x%02x.\n",numInterfaces));

 WDF_USB_DEVICE_SELECT_CONFIG_PARAMS_INIT_SINGLE_INTERFACE(&configParams);


 //     USB
 status=WdfUsbTargetDeviceSelectConfig(pDeviceContext->UsbDevice,WDF_NO_OBJECT_ATTRIBUTES,&configParams);
 if(!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_iAP: WdfUsbTargetDeviceSelectConfig failed 0x%x\n", status));
  return(status);
 }
 pDeviceContext->UsbInterface=configParams.Types.SingleInterface.ConfiguredUsbInterface;
 // 2 :     
 // 2 :     
 pDeviceContext->BulkReadPipe=NULL;
 pDeviceContext->BulkWritePipe=NULL;
 WDF_USB_PIPE_INFORMATION_INIT(&pipeConfig);
 while(1)
 {
  pipe=WdfUsbInterfaceGetConfiguredPipe(pDeviceContext->UsbInterface,ep_index,&pipeConfig);
  if (pipe==NULL) break;//   
  KdPrint(("FlirOneGen2_iAP: EndPoint addr:0x%02x\n",pipeConfig.EndpointAddress));
  if (pipeConfig.EndpointAddress==BULK_READ_ENDPOINT_ADDR) pDeviceContext->BulkReadPipe=pipe;
  if (pipeConfig.EndpointAddress==BULK_WRITE_ENDPOINT_ADDR) pDeviceContext->BulkWritePipe=pipe;
  ep_index++;
 }

 //  
 if (pDeviceContext->BulkReadPipe!=NULL)
 {
  KdPrint(("FlirOneGen2_iAP: BulkReadPipe is OK.\n"));
  WdfUsbTargetPipeSetNoMaximumPacketSizeCheck(pDeviceContext->BulkReadPipe);//      
 }
 else KdPrint(("FlirOneGen2_iAP: BulkReadPipe is NULL!\n"));
 
 //  
 if (pDeviceContext->BulkWritePipe!=NULL)
 {
  KdPrint(("FlirOneGen2_iAP: BulkWritePipe is OK.\n"));
  WdfUsbTargetPipeSetNoMaximumPacketSizeCheck(pDeviceContext->BulkWritePipe);//     
 }
 else KdPrint(("FlirOneGen2_iAP: BulkWritePipe is NULL!\n"));
 return(status);
}
 
//----------------------------------------------------------------------------------------------------
//,   DeviceIoControl
//----------------------------------------------------------------------------------------------------
VOID EvtIoDeviceControl(IN WDFQUEUE Queue,IN WDFREQUEST Request,IN size_t OutputBufferLength, IN size_t InputBufferLength,IN ULONG IoControlCode)
{
 WDFDEVICE device;//
 PDEVICE_CONTEXT pDeviceContext;// 
 NTSTATUS status;// 
 WDF_USB_CONTROL_SETUP_PACKET controlSetupPacket;// 
 WDF_MEMORY_DESCRIPTOR memDesc;//  
 WDFMEMORY memory;// 
 WDF_REQUEST_SEND_OPTIONS sendOptions;//   
 size_t bytesTransferred=0;//  
 unsigned char data[2]={0,0};
 long value=0;
 long index=0;
 unsigned char ok=0;
 long n;
 unsigned char *ptr;
  
 UNREFERENCED_PARAMETER(InputBufferLength);
 UNREFERENCED_PARAMETER(OutputBufferLength);

 KdPrint(("FlirOneGen2_iAP: EvtIoDeviceControl.\n"));//     

 status=STATUS_INVALID_DEVICE_REQUEST;//  
  
 device=WdfIoQueueGetDevice(Queue);
 pDeviceContext=GetDeviceContext(device);
  
 if (IoControlCode==IOCTL_FLIR_ONE_GEN2_FRAME_STOP)// 
 {
  KdPrint(("FlirOneGen2_iAP: IOCTL_FLIR_ONE_GEN2_FRAME_STOP.\n"));//     
  value=0;
  index=2;
  ok=1;
 }
 if (IoControlCode==IOCTL_FLIR_ONE_GEN2_FRAME_START)// 
 {
  KdPrint(("FlirOneGen2_iAP: IOCTL_FLIR_ONE_GEN2_FRAME_START.\n"));//     
  value=1;
  index=2;
  ok=1;
 }
 if (IoControlCode==IOCTL_FLIR_ONE_GEN2_FILEIO_STOP)// 
 {
  KdPrint(("FlirOneGen2_iAP: IOCTL_FLIR_ONE_GEN2_FILEIO_STOP.\n"));//     
  value=0;
  index=1;
  ok=1;
 }
 if (IoControlCode==IOCTL_FLIR_ONE_GEN2_FILEIO_START)// 
 {
  KdPrint(("FlirOneGen2_iAP: IOCTL_FLIR_ONE_GEN2_FILEIO_START.\n"));//     
  value=1;
  index=1;
  ok=1;
 }
 if (ok==1) 
 {
  //      
  //  
  status=WdfRequestRetrieveInputMemory(Request,&memory);
  if (!NT_SUCCESS(status)) 
  {
   KdPrint(("FlirOneGen2_iAP: WdfRequestRetrieveMemory failed 0x%x",status));
   WdfRequestCompleteWithInformation(Request,status,bytesTransferred);
   return;
  }
  //    

/* Flir config
            01 0b 01 00 01 00 00 00 c4 d5
            0 bmRequestType = 01
            1 bRequest = 0b
            2 wValue 0001 type (H) index (L)    stop=0/start=1 (Alternate Setting)
            4 wIndex 01                         interface 1/2
            5 wLength 00
            6 Data 00 00
			*/

  //WDF_USB_CONTROL_SETUP_PACKET_INIT_VENDOR(&controlSetupPacket,BmRequestHostToDevice,BmRequestToInterface,USB_REQUEST_SET_INTERFACE,value,index);
  WDF_USB_CONTROL_SETUP_PACKET_INIT(&controlSetupPacket,BmRequestHostToDevice,BmRequestToInterface,USB_REQUEST_SET_INTERFACE,value,index);  
  KdPrint(("FlirOneGen2_iAP: Send: "));
  ptr=(unsigned char*)(&controlSetupPacket);
  for(n=0;n<sizeof(controlSetupPacket);n++)
  {
   KdPrint(("0x%02x ",ptr[n]));
  }
  KdPrint(("\n"));  

  WDF_MEMORY_DESCRIPTOR_INIT_HANDLE(&memDesc,memory,NULL);
  WDF_MEMORY_DESCRIPTOR_INIT_BUFFER(&memDesc,data,0);

  //  
  WDF_REQUEST_SEND_OPTIONS_INIT(&sendOptions,WDF_REQUEST_SEND_OPTION_TIMEOUT);
  WDF_REQUEST_SEND_OPTIONS_SET_TIMEOUT(&sendOptions,WDF_REL_TIMEOUT_IN_MS(200));  
  
  status=WdfUsbTargetDeviceSendControlTransferSynchronously(pDeviceContext->UsbDevice,WDF_NO_HANDLE,&sendOptions,&controlSetupPacket,&memDesc,(PULONG)&bytesTransferred);
  if (!NT_SUCCESS(status)) 
  {
   KdPrint(("FlirOneGen2_iAP: SendControlTransfer failed 0x%x", status));
   WdfRequestCompleteWithInformation(Request,status,bytesTransferred);
   return;
  }
  KdPrint(("FlirOneGen2_iAP: Ok. Transferred:0x%02x\n",bytesTransferred));//     
 }
 WdfRequestCompleteWithInformation(Request,status,bytesTransferred);
 return;
}
 
//----------------------------------------------------------------------------------------------------
//  
//Queue-.
//Request- .
//Length-   
//----------------------------------------------------------------------------------------------------
VOID EvtIoRead(IN WDFQUEUE Queue,IN WDFREQUEST Request,IN size_t Length)
{
 NTSTATUS status;// 
 WDFUSBPIPE pipe;//
 WDFMEMORY reqMemory;
 PDEVICE_CONTEXT pDeviceContext;// 
 BOOLEAN ret;
  
 UNREFERENCED_PARAMETER(Length);//,       

 KdPrint(("FlirOneGen2_iAP: EvtIoRead.\n"));//     
 
 pDeviceContext=GetDeviceContext(WdfIoQueueGetDevice(Queue));//  
  
 pipe=pDeviceContext->BulkReadPipe;//   
 if (pipe==NULL)
 {
  KdPrint(("FlirOneGen2_iAP: Pipe is NULL!\n"));//     
  status=STATUS_SUCCESS;
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
 
 status=WdfRequestRetrieveOutputMemory(Request,&reqMemory);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_iAP: WdfRequestRetrieveOutputMemory failed 0x%x", status));
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
 //   
 status=WdfUsbTargetPipeFormatRequestForRead(pipe,Request,reqMemory,NULL);
 if (!NT_SUCCESS(status))// 
 {
  KdPrint(("FlirOneGen2_iAP: WdfUsbTargetPipeFormatRequestForRead failed 0x%x", status));
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
 //  
 WdfRequestSetCompletionRoutine(Request,EvtRequestReadCompletionRoutine,pipe);
  
 ret=WdfRequestSend(Request,WdfUsbTargetPipeGetIoTarget(pipe),WDF_NO_SEND_OPTIONS);
 if (ret==FALSE) 
 {
  KdPrint(("FlirOneGen2_iAP: WdfRequestSend failed 0x%x", status));
  status=WdfRequestGetStatus(Request);
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return; 
 } 
 else return;
 KdPrint(("FlirOneGen2_iAP: Ok."));
 WdfRequestCompleteWithInformation(Request,status,0);// 
 return;
}
 
//----------------------------------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------------------------------
VOID EvtRequestReadCompletionRoutine(IN WDFREQUEST Request,IN WDFIOTARGET Target,PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,IN WDFCONTEXT Context)
{
 NTSTATUS status;// 
 size_t bytesRead=0;//  
 PWDF_USB_REQUEST_COMPLETION_PARAMS usbCompletionParams;
  
 UNREFERENCED_PARAMETER(Target);//,       
 UNREFERENCED_PARAMETER(Context);//,       
  
 status=CompletionParams->IoStatus.Status;
  
 usbCompletionParams=CompletionParams->Parameters.Usb.Completion;
 bytesRead=usbCompletionParams->Parameters.PipeRead.Length;
  
 if (NT_SUCCESS(status))//  
 {
  KdPrint(("FlirOneGen2_iAP: Number of bytes read: %I64d\n",(INT64)bytesRead));
 }
 else 
 {
  KdPrint(("FlirOneGen2_iAP: Read failed - request status 0x%x UsbdStatus 0x%x\n",status,usbCompletionParams->UsbdStatus));
 }
 WdfRequestCompleteWithInformation(Request,status,bytesRead);// 
 return;
}
//----------------------------------------------------------------------------------------------------
//  
//Queue-.
//Request- .
//Length-   
//----------------------------------------------------------------------------------------------------
VOID EvtIoWrite(IN WDFQUEUE Queue,IN WDFREQUEST Request,IN size_t Length)
{
 NTSTATUS status;// 
 WDFUSBPIPE pipe;//
 WDFMEMORY reqMemory;
 PDEVICE_CONTEXT pDeviceContext;// 
 BOOLEAN ret;
  
 UNREFERENCED_PARAMETER(Length);//,       
  
 pDeviceContext=GetDeviceContext(WdfIoQueueGetDevice(Queue));//  
  
 pipe=pDeviceContext->BulkWritePipe;//   
  
 status=WdfRequestRetrieveInputMemory(Request,&reqMemory);
 if (!NT_SUCCESS(status))// 
 {
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
 //  
 status=WdfUsbTargetPipeFormatRequestForWrite(pipe,Request,reqMemory,NULL);
 if (!NT_SUCCESS(status))// 
 {
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 }
  
 WdfRequestSetCompletionRoutine(Request,EvtRequestWriteCompletionRoutine,pipe);
  
 ret=WdfRequestSend(Request,WdfUsbTargetPipeGetIoTarget(pipe),WDF_NO_SEND_OPTIONS);
 if (ret==FALSE)//
 {
  status=WdfRequestGetStatus(Request);
  WdfRequestCompleteWithInformation(Request,status,0);// 
  return;
 } 
 else return;
  
 WdfRequestCompleteWithInformation(Request,status,0);// 
 return;
}
//----------------------------------------------------------------------------------------------------
//  
//----------------------------------------------------------------------------------------------------
VOID EvtRequestWriteCompletionRoutine(IN WDFREQUEST Request,IN WDFIOTARGET Target,PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,IN WDFCONTEXT Context)
{
 NTSTATUS status;
 size_t bytesWritten=0;//  
 PWDF_USB_REQUEST_COMPLETION_PARAMS usbCompletionParams;
  
 UNREFERENCED_PARAMETER(Target);//,       
 UNREFERENCED_PARAMETER(Context);//,       
  
 status=CompletionParams->IoStatus.Status;
  
 usbCompletionParams=CompletionParams->Parameters.Usb.Completion;
  
 bytesWritten=usbCompletionParams->Parameters.PipeWrite.Length;
 if (NT_SUCCESS(status))//  
 {
  KdPrint(("FlirOneGen2_iAP: Number of bytes written: %I64d\n",(INT64)bytesWritten));
 }
 else// 
 {
  KdPrint(("FlirOneGen2_iAP: Write failed: request Status 0x%x UsbdStatus 0x%x\n",
  status, usbCompletionParams->UsbdStatus));
 }
 WdfRequestCompleteWithInformation(Request,status,bytesWritten);// 
 return;
}
